from PyQt5.QtCore import QPointF,QRectF,Qt
from PyQt5.QtGui import QBrush,QColor,QPen,QPolygonF
from qgis.core import QgsCoordinateTransform,QgsMapToPixel,QgsPointXY,QgsSimpleLineSymbolLayer,QgsSymbolRenderContext
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject

# Class MapObjectSymbolLayer implements a (marker) symbol layer that draws MapObjects on the QGIS map using QPainter
#
class MapObjectSymbolLayer(QgsSimpleLineSymbolLayer):
    def __init__(self):
        QgsSimpleLineSymbolLayer.__init__(self)
        self.mapObject = None
        self.mapObjectIndex = 0

    def layerType(self):
        return "MapObject"

    def properties(self):
        return {}
 
    def clone(self):
        return MapObjectSymbolLayer()
 
    def startRender(self,context: QgsSymbolRenderContext):
        pass
 
    def stopRender(self,context: QgsSymbolRenderContext):
        pass

    # Called by the MapObjectRenderer before rendering a certain MapObject feature
    #
    def setMapObject(self,mapObject: MapObject,mapObjectIndex: int):
        self.mapObject = mapObject
        self.mapObjectIndex = mapObjectIndex
 
    # Called by QGIS to render a MapObject point feature
    #
    def renderPoint(self,point,context: QgsSymbolRenderContext):
        self.renderMapObject(context)
 
    # Called by QGIS to render a MapObject polyline feature
    #
    def renderPolyline(self,points,context: QgsSymbolRenderContext):
        self.renderMapObject(context)
 
    # Called by QGIS to render a MapObject polygon feature
    #
    def renderPolygon(self,outline,rings,context: QgsSymbolRenderContext):
        self.renderMapObject(context)
 
    # Render the MapObject feature
    #
    def renderMapObject(self,context: QgsSymbolRenderContext):
        if (self.mapObject==None):
            Message.print(__name__,"No mapObject to render")
            return
        painter = context.renderContext().painter() # QPainter
        # get the transformations to draw objects on the QGIS map
        worldToMap = context.renderContext().coordinateTransform()
        mapToPixel = context.renderContext().mapToPixel()
        # calculate the position of the label as centroid (average) of the geometry points projected on the map
        labelX = 0.0
        labelY = 0.0
        labelN = 0
        # do we have an area to render?
        if (self.mapObject.spatialArea!=None):
            pen = QPen(QColor(self.mapObject.color.red,self.mapObject.color.green,self.mapObject.color.blue,self.mapObject.color.alpha))
            pen.setWidthF(self.mapObject.lineWidth)
            painter.setPen(pen)
            if (self.mapObject.lineWidth<=0.0): painter.setPen(Qt.NoPen)
            brush = QBrush(QColor(self.mapObject.fillColor.red,self.mapObject.fillColor.green,self.mapObject.fillColor.blue,self.mapObject.fillColor.alpha))
            painter.setBrush(brush)
            for j in range(0,self.mapObject.spatialArea.parts.size()):
                part = self.mapObject.spatialArea.parts.get(j)
                polygon = QPolygonF()
                for k in range(0,part.points.size()):
                    point = part.points.get(k)
                    pixel = mapToPixel.transform(worldToMap.transform(QgsPointXY(point.x,point.y)))
                    polygon.append(QPointF(pixel.x(),pixel.y()))
                    labelX += pixel.x()
                    labelY += pixel.y()
                    labelN += 1
                if (part.points.size()>2):
                    painter.drawPolygon(polygon)
                if (part.points.size()==2):
                    painter.drawPolyline(polygon)
        # do we have a line to render?
        if (self.mapObject.spatialLine!=None):
            pen = QPen(QColor(self.mapObject.color.red,self.mapObject.color.green,self.mapObject.color.blue,self.mapObject.color.alpha))
            pen.setWidthF(self.mapObject.lineWidth)
            painter.setPen(pen)
            painter.setBrush(Qt.NoBrush)
            for j in range(0,self.mapObject.spatialLine.parts.size()):
                part = self.mapObject.spatialLine.parts.get(j)
                polygon = QPolygonF()
                for k in range(0,part.points.size()):
                    point = part.points.get(k)
                    pixel = mapToPixel.transform(worldToMap.transform(QgsPointXY(point.x,point.y)))
                    polygon.append(QPointF(pixel.x(),pixel.y()))
                    labelX += pixel.x()
                    labelY += pixel.y()
                    labelN += 1
                if (part.points.size()>1):
                    painter.drawPolyline(polygon)
        # do we have a point to render?
        if (self.mapObject.spatialPoint!=None) and (self.mapObject.symbolPath==None):
            pixel = mapToPixel.transform(worldToMap.transform(QgsPointXY(self.mapObject.spatialPoint.x,self.mapObject.spatialPoint.y)))
            symbolName = self.mapObject.symbolName
            symbolSize = self.mapObject.symbolSize
            pen = QPen(QColor(self.mapObject.color.red,self.mapObject.color.green,self.mapObject.color.blue,self.mapObject.color.alpha))
            pen.setWidthF(self.mapObject.lineWidth)
            painter.setPen(pen)
            if (self.mapObject.lineWidth<=0.0): painter.setPen(Qt.NoPen)
            brush = QBrush(QColor(self.mapObject.fillColor.red,self.mapObject.fillColor.green,self.mapObject.fillColor.blue,self.mapObject.fillColor.alpha))
            painter.setBrush(brush)
            symbolX = pixel.x()-0.5*symbolSize
            symbolY = pixel.y()-0.5*symbolSize
            if (symbolName==None): symbolName = "circle_filled"
            if (symbolName=="circle_filled"):
                painter.drawEllipse(QRectF(symbolX,symbolY,symbolSize,symbolSize))
            elif (symbolName=="square_filled"):
                painter.drawRect(QRectF(symbolX,symbolY,symbolSize,symbolSize))
            else:
                painter.drawRect(QRectF(symbolX,symbolY,symbolSize,symbolSize))
            labelX += pixel.x()
            labelY += pixel.y()
            labelN += 1
        # do we have a symbol to render?
        if (self.mapObject.spatialPoint!=None) and (self.mapObject.symbolPath!=None):
            pen = QPen(QColor(self.mapObject.color.red,self.mapObject.color.green,self.mapObject.color.blue,self.mapObject.color.alpha))
            pen.setWidthF(self.mapObject.lineWidth)
            painter.setPen(pen)
            brush = QBrush(QColor(self.mapObject.fillColor.red,self.mapObject.fillColor.green,self.mapObject.fillColor.blue,self.mapObject.fillColor.alpha))
            painter.setBrush(brush)
            pixelOrigin = mapToPixel.transform(worldToMap.transform(QgsPointXY(self.mapObject.spatialPoint.x,self.mapObject.spatialPoint.y)))
            labelX += pixelOrigin.x()
            labelY += pixelOrigin.y()
            labelN += 1
            for j in range(0,self.mapObject.symbolPath.parts.size()):
                part = self.mapObject.symbolPath.parts.get(j)
                polygon = QPolygonF()
                for k in range(0,part.points.size()):
                    pixel = part.points.get(k)
                    polygon.append(QPointF(pixelOrigin.x()+pixel.x,pixelOrigin.y()+pixel.y))
                if (part.points.size()>2):
                    painter.drawPolygon(polygon)
                if (part.points.size()==2):
                    painter.drawPolyline(polygon)
        # do we have a label to render?
        if (self.mapObject.label!=None) and (labelN>0):
            labelText = self.mapObject.label
            labelBounds = painter.fontMetrics().boundingRect(labelText)
            labelPosX = (labelX/labelN)-0.5*labelBounds.width()
            labelPosY = (labelY/labelN)+0.5*labelBounds.height()
            if (self.mapObject.labelPosition!=None):
                labelPosX += self.mapObject.labelPosition.x
                labelPosY += self.mapObject.labelPosition.y
            labelColor = QColor(self.mapObject.labelColor.red,self.mapObject.labelColor.green,self.mapObject.labelColor.blue,self.mapObject.labelColor.alpha)
            painter.setPen(labelColor)
            painter.drawText(int(labelPosX),int(labelPosY),labelText)
