from pathlib import Path
from PyQt5.QtCore import QTimer
from qgis.core import QgsApplication,QgsCoordinateReferenceSystem,QgsCoordinateTransform,QgsFeature,QgsPointXY,QgsProject,QgsRectangle,QgsTask,QgsVectorLayer,QgsWkbTypes
from qgis.gui import QgisInterface,QgsMapToolEmitPoint
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.AMap import AMap
from ..orbitgt.mapping3d.sdk.viewer.Bounds import Bounds
from ..orbitgt.mapping3d.sdk.viewer.Collections import Collections
from ..orbitgt.mapping3d.sdk.viewer.Color import Color
from ..orbitgt.mapping3d.sdk.viewer.Constants import Constants
from ..orbitgt.mapping3d.sdk.viewer.Coordinate import Coordinate
from ..orbitgt.mapping3d.sdk.viewer.CoordinatePart import CoordinatePart
from ..orbitgt.mapping3d.sdk.viewer.CoordinatePath import CoordinatePath
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from ..orbitgt.mapping3d.sdk.viewer.SDKViewerHost import SDKViewerHost
from ..orbitgt.mapping3d.sdk.viewer.Viewer3DM import Viewer3DM
from .MapObjectLayer import MapObjectLayer
from .MapObjects import MapObjects
from .Overlays import Overlays
from .UpdateFieldOfView import UpdateFieldOfView
from .UpdateFootprint import UpdateFootprint
from .UpdateGlobal import UpdateGlobal
from .UpdateMeasurement import UpdateMeasurement

# Class PluginQGIS implements the 3D Mapping Viewer plugin for QGIS
#
# Copyright Orbit Geospatial Technologies
#
class PluginQGIS:

    # Create a new QGIS plugin instance
    # QGIS Python API: https://qgis.org/pyqgis/3.0/
    #
    def __init__(self,iface: QgisInterface,pluginDir: str):
        Message.print(__name__,"Creating QGIS plugin")
        self.iface = iface
        self.pluginDir = pluginDir
        # Initialize
        self.pickPositionTool = None
        self.pickPositionFocus = False
        # initialize the 3D Mapping viewer (starts the python viewer host and the websocket server)
        # we pass the auto-login application state (remember-me) and coordinate system preference to the startup options
        self.viewerHost = SDKViewerHost()
        viewerOptions = AMap[object](Collections.TYPE_OBJECT)
        appState = self.readApplicationStateFile();
        if (appState!=None): viewerOptions.set(Constants.STARTUP_APP_STATE,appState)
        viewerOptions.set(Constants.STARTUP_CRS,MapObjects.CRS)
        self.viewerHost.initializeViewer("QGIS 3D Mapping plugin",viewerOptions,self.lostConnection)
        # listen to various signals emitted by the 3D Mapping viewer
        self.getViewer().onAppStateChanged.add(self.onAppStateChanged)
        self.getViewer().onFootprintsChanged.add(self.onFootprintsChanged)
        self.getViewer().onFieldOfViewsChanged.add(self.onFieldOfViewsChanged)
        self.getViewer().onMeasureStateChanged.add(self.onMeasureStateChanged)
        self.getViewer().onGlobalsChanged.add(self.onGlobalsChanged)
        self.getViewer().onZoomToExtent.add(self.onZoomToExtent)
        self.getViewer().onPickPositionRequested.add(self.onPickPositionRequested)
        self.getViewer().onPickPositionCancelled.add(self.onPickPositionCancelled)
        # start the viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
        self.viewerHost.startViewer()
        # create the layers
        self.footprintLayer = MapObjectLayer("3D Mapping Footprints",self.iface)
        self.fieldOfViewLayer = MapObjectLayer("3D Mapping Field of Views",self.iface)
        self.measureLayer = MapObjectLayer("3D Mapping Measurements",self.iface)
        self.globalLayer = MapObjectLayer("3D Mapping Globals",self.iface)
        # create the updaters for the layers
        self.footprintUpdater = UpdateFootprint(self.getViewer(),self.footprintLayer,self.iface)
        self.fieldOfViewUpdater = UpdateFieldOfView(self.getViewer(),self.fieldOfViewLayer,self.iface)
        self.measurementUpdater = UpdateMeasurement(self.getViewer(),self.measureLayer,self.iface)
        self.globalUpdater = UpdateGlobal(self.getViewer(),self.globalLayer,self.iface)
        # start the refresh timer
        self.refreshTimer = QTimer()
        self.refreshTimer.timeout.connect(self.doRefresh)
        self.refreshTimer.start(250) # milliseconds
        # listen to changes in the mapcanvas extent
        self.iface.mapCanvas().extentsChanged.connect(self.updatedMapCanvas)
        # print the 'home' path of the user (where the application state is stored)
        Message.print(__name__,"The user home path is '"+str(Path.home())+"'")
        
    # Close the plugin
    #
    def close(self) -> None:
        try:
            if (self.iface==None): return
            self.refreshTimer.stop()
            self.closeLayers()
            self.iface.mapCanvas().extentsChanged.disconnect(self.updatedMapCanvas)
            self.iface = None
        except Exception as e:
            Message.printError(__name__,e,"Error while closing the plugin")

    # Close the map layers (should be called from the main UI thread)
    #
    def closeLayers(self) -> None:
        Message.print(__name__,"Closing layers")
        self.footprintLayer.unload()
        self.fieldOfViewLayer.unload()
        self.measureLayer.unload()
        self.globalLayer.unload()
    def closeLayers0(self,task: QgsTask) -> None:
        pass
    def closeLayers1(self,taskException) -> None:
        self.closeLayers()
        self.iface.mapCanvas().refresh()
    def closeLayers2(self) -> None:
        self.closeLayersTask = QgsTask.fromFunction("Closing map layers",self.closeLayers0,on_finished=self.closeLayers1)
        QgsApplication.taskManager().addTask(self.closeLayersTask)
    
    # Get the 3D Mapping viewer
    #
    def getViewer(self) -> Viewer3DM:
        return self.viewerHost.getViewer()
    
    # Read the application state file from disk
    #
    def readApplicationStateFile(self) -> str:
        try:
            fileName = str(Path.home())+"/qgis_3dmc_app_state.ini"
            file = open(fileName,"r")
            appState = file.readline()
            file.close()
            Message.print(__name__,"Read application state from file '"+fileName+"'")
            return appState
        except FileNotFoundError:
            return None
        except Exception as e:
            Message.printError(__name__,e,"Error while reading application state")
            return None
    
    # Write the application state file to disk
    #
    def writeApplicationStateFile(self,appState: str) -> None:
        try:
            fileName = str(Path.home())+"/qgis_3dmc_app_state.ini"
            file = open(fileName,"w") 
            file.write(appState) 
            file.close() 
            Message.print(__name__,"Written application state to file '"+fileName+"'")
        except Exception as e:
            Message.printError(__name__,e,"Error while writing application state")
            return None
        
    # Signal called by the 3D Mapping viewer when the application (login) state has changed. We save the new state to disk to enable auto-login (remember-me) for the next day
    #
    def onAppStateChanged(self,appState: str) -> None:
        Message.print(__name__,"Got 'onAppStateChanged' signal ("+appState+")")
        self.writeApplicationStateFile(appState)
    
    # Signal called by the 3D Mapping viewer the footprints have changed
    #
    def onFootprintsChanged(self,message: str) -> None:
        Message.print(__name__,"Got 'onFootprintsChanged' signal")
        # request new footprints
        self.footprintUpdater.updatedMapCanvas()
    
    # Signal called by the 3D Mapping viewer when the field-of-views have changed
    #
    def onFieldOfViewsChanged(self,message: str) -> None:
        Message.print(__name__,"Got 'onFieldOfViewsChanged' signal")
        # request new field-of-views
        self.fieldOfViewUpdater.updatedViews()
    
    # Signal called by the 3D Mapping viewer when the measure state has changed
    #
    def onMeasureStateChanged(self,message: str) -> None:
        Message.print(__name__,"Got 'onMeasureStateChanged' signal")
        # request a new measure state
        self.measurementUpdater.updatedMeasurement()
    
    # Signal called by the 3D Mapping viewer when the global map objects have changed
    #
    def onGlobalsChanged(self,message: str) -> None:
        Message.print(__name__,"Got 'onGlobalsChanged' signal")
        # request new globals
        self.globalUpdater.updatedGlobals()
    
    # Signal called by the 3D Mapping viewer when a zoomout to a spatial extent is requested by the user (zoom-to-extent step 1)
    #
    def onZoomToExtent(self,worldBounds: Bounds) -> None:
        Message.print(__name__,"Got 'onZoomToExtent' signal ("+worldBounds.toString()+")")
        # Create a task to run the 'setExtent' method in the main UI thread
        self.zoomToExtentBounds = worldBounds
        self.zoomToExtentTask = QgsTask.fromFunction("Zoom to extent",self.onZoomToExtent2,on_finished=self.onZoomToExtent3)
        QgsApplication.taskManager().addTask(self.zoomToExtentTask)

    # Do the zoom-to-extent procedure (called in background thread) (zoom-to-extent step 2)
    #
    def onZoomToExtent2(self,task: QgsTask):
        pass

    # Finish the zoom-to-extent procedure (called in main UI thread) (zoom-to-extent step 3)
    #
    def onZoomToExtent3(self,taskException):
        try:
            # let QGIS zoom the map canvas to the given bounds
            worldBounds = self.zoomToExtentBounds
            worldToMap = QgsCoordinateTransform(QgsCoordinateReferenceSystem("EPSG:"+MapObjects.CRS),self.iface.mapCanvas().mapSettings().destinationCrs(),QgsProject.instance())
            mapBounds = QgsRectangle(worldToMap.transform(QgsPointXY(worldBounds.minX,worldBounds.minY)),worldToMap.transform(QgsPointXY(worldBounds.maxX,worldBounds.maxY)))
            mapBounds.include(worldToMap.transform(QgsPointXY(worldBounds.maxX,worldBounds.minY)))
            mapBounds.include(worldToMap.transform(QgsPointXY(worldBounds.minX,worldBounds.maxY)))
            self.iface.mapCanvas().setExtent(mapBounds)
            self.iface.mapCanvas().refresh()
        except Exception as e:
            Message.printError(__name__,e,"Error while during zoom-to-extent")

    # Signal called by the 3D Mapping viewer when the viewer needs a location to be picked on the reference map by the user (pick-position step 1)
    #
    def onPickPositionRequested(self,message: str) -> None:
        Message.print(__name__,"Got 'onPickPositionRequested' signal")
        # Create a task to run the 'setMapTool' method in the main UI thread
        self.pickPositionFocus = False
        self.activatePickPositionTask = QgsTask.fromFunction("Activate the pick position tool",self.onPickPositionRequested2,on_finished=self.onPickPositionRequested3)
        QgsApplication.taskManager().addTask(self.activatePickPositionTask)

    # Activate the pick-position tool (called in background thread) (pick-position step 2)
    #
    def onPickPositionRequested2(self,task: QgsTask):
        pass

    # Activate the pick-position tool (called in main UI thread) (pick-position step 3)
    #
    def onPickPositionRequested3(self,taskException):
        try:
            if (self.pickPositionTool==None):
                Message.print(__name__,"Creating the pick-position tool")
                self.pickPositionTool = QgsMapToolEmitPoint(self.iface.mapCanvas())
                self.pickPositionTool.canvasClicked.connect(self.onPickPositionRequested4)
            Message.print(__name__,"Activating the pick-position tool")
            self.iface.mapCanvas().setMapTool(self.pickPositionTool)
        except Exception as e:
            Message.printError(__name__,e,"Error while activating the pick-position tool")

    # Called when the user clicked on the map canvas using the pick-position tool (pick-position step 4)
    #
    def onPickPositionRequested4(self,mapPoint: QgsPointXY,button) -> None:
        try:
            Message.print(__name__,"Clicked map position ("+str(mapPoint.x())+","+str(mapPoint.y())+") on the map")
            mapToWorld = QgsCoordinateTransform(self.iface.mapCanvas().mapSettings().destinationCrs(),QgsCoordinateReferenceSystem("EPSG:"+MapObjects.CRS),QgsProject.instance())
            worldPoint = mapToWorld.transform(mapPoint)
            Message.print(__name__,"Sending position ("+str(worldPoint.x())+","+str(worldPoint.y())+") ("+MapObjects.CRS+") to the viewer")
            position = Coordinate.fromXY(MapObjects.CRS,worldPoint.x(),worldPoint.y())
            if (self.pickPositionFocus): self.getViewer().setFocusPosition(position)
            else: self.getViewer().setPickPosition(position)
            self.pickPositionFocus = False
            # active the standard pan function
            self.iface.actionPan().trigger()
        except Exception as e:
            Message.printError(__name__,e,"Error while sending clicked position to viewer")

    # Signal called by the 3D Mapping viewer when the pick-position operation needs to be cancelled (cancel pick-position step 1)
    #
    def onPickPositionCancelled(self,message: str) -> None:
        Message.print(__name__,"Got 'onPickPositionCancelled' signal")
        # Create a task to run the 'setMapTool' method in the main UI thread
        self.cancelPickPositionTask = QgsTask.fromFunction("Cancel the pick position tool",self.onPickPositionCancelled2,on_finished=self.onPickPositionCancelled3)
        QgsApplication.taskManager().addTask(self.cancelPickPositionTask)

    # Cancel the pick-position tool (called in background thread) (cancel pick-position step 2)
    #
    def onPickPositionCancelled2(self,task: QgsTask):
        pass

    # Cancel the pick-position tool (called in main UI thread) (cancel pick-position step 3)
    #
    def onPickPositionCancelled3(self,taskException):
        # active the standard pan function
        self.iface.actionPan().trigger()

    # Called by QGIS when the mapcanvas extent has changed
    #
    def updatedMapCanvas(self):
        # request new footprints
        Message.print(__name__,"Updated mapcanvas extent")
        self.footprintUpdater.updatedMapCanvas()
    
    # Called by the refresh timer at a regular interval
    #
    def doRefresh(self):
        if (self.viewerHost.isViewerReady()):
            if (self.viewerHost.hasViewerConnection()):
                self.footprintUpdater.checkForUpdates()
                self.fieldOfViewUpdater.checkForUpdates()
                self.measurementUpdater.checkForUpdates()
                self.globalUpdater.checkForUpdates()

    # Called when the user pressed the 'start' button on the toolbar again
    def restart(self):
        # start a new viewer (which opens a javascript browser window that connects to this plugin instance using websockets)
        if (self.viewerHost.hasViewerConnection()==False): self.viewerHost.startViewer()
        # update all map objects
        self.footprintUpdater.updatedMapCanvas()
        self.fieldOfViewUpdater.updatedViews()
        self.measurementUpdater.updatedMeasurement()
        self.globalUpdater.updatedGlobals()

    # Toolbar action callback method
    #
    def doShowFootprints(self,visible: bool):
        if (self.viewerHost.isViewerReady()):
            self.footprintLayer.setVisible(visible)
            self.fieldOfViewLayer.setVisible(visible)
            self.measureLayer.setVisible(visible)
            self.globalLayer.setVisible(visible)
            self.iface.mapCanvas().refreshAllLayers()

    # Toolbar action callback method
    #
    def doPickLocation(self):
        if (self.viewerHost.isViewerReady()):
            self.pickPositionFocus = True
            self.onPickPositionRequested3(None);

    # Toolbar action callback method
    #
    def doSaveMeasurement(self):
        try:
            Message.print(__name__,"doSaveMeasurement")
            if (self.viewerHost.isViewerReady()):
                self.measurementUpdater.doSaveMeasurement()
        except Exception as e:
            Message.printError(__name__,e,"Error while saving measurement")
        
    # Toolbar action callback method
    #
    def doAddOverlays(self):
        try:
            Message.print(__name__,"doAddOverlays")
            if (self.viewerHost.isViewerReady()):
                Overlays.doAddOverlays(self.iface,self.getViewer())
        except Exception as e:
            Message.printError(__name__,e,"Error while adding overlays")
    
    # Called when the connection to the javascript browser has been lost (when the user closes the browser tab)
    #
    def lostConnection(self) -> None:
        try:
            self.closeLayers2()
        except Exception as e:
            Message.printError(__name__,e,"Error while handling lost-connection")
